@@ -249,7 +249,7 @@ static int osmo_sock_init_tail(int fd, uint16_t type, unsigned int flags)
 			if (rc < 0) {
 				LOGP(DLGLOBAL, LOGL_ERROR, "unable to listen on socket: %s\n",
 					strerror(errno));
-				return rc;
+				return -errno;
 			}
 			break;
 		}
@@ -561,19 +561,21 @@ int osmo_sock_init_osa(uint16_t type, uint8_t proto,
 			rc = setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR,
 					&on, sizeof(on));
 			if (rc < 0) {
+				int err = errno;
 				_SOCKADDR_TO_STR(sastr, local);
 				LOGP(DLGLOBAL, LOGL_ERROR,
 				     "cannot setsockopt socket: " OSMO_SOCKADDR_STR_FMT ": %s\n",
-				     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(errno));
+				     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(err));
 				close(sfd);
 				return rc;
 			}
 		}
 
 		if (bind(sfd, &local->u.sa, sizeof(struct osmo_sockaddr)) == -1) {
+			int err = errno;
 			_SOCKADDR_TO_STR(sastr, local);
 			LOGP(DLGLOBAL, LOGL_ERROR, "unable to bind socket: " OSMO_SOCKADDR_STR_FMT ": %s\n",
-				     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(errno));
+				     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(err));
 			close(sfd);
 			return -1;
 		}
@@ -594,9 +596,10 @@ int osmo_sock_init_osa(uint16_t type, uint8_t proto,
 
 		rc = connect(sfd, &remote->u.sa, sizeof(struct osmo_sockaddr));
 		if (rc != 0 && errno != EINPROGRESS) {
+			int err = errno;
 			_SOCKADDR_TO_STR(sastr, remote);
 			LOGP(DLGLOBAL, LOGL_ERROR, "unable to connect socket: " OSMO_SOCKADDR_STR_FMT ": %s\n",
-			     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(errno));
+			     OSMO_SOCKADDR_STR_FMT_ARGS(sastr), strerror(err));
 			close(sfd);
 			return rc;
 		}
@@ -693,6 +696,76 @@ static int addrinfo_to_sockaddr(uint16_t family, const struct addrinfo **result,
 	return 0;
 }
 
+static int setsockopt_sctp_auth_supported(int fd, uint32_t val)
+{
+#ifdef SCTP_AUTH_SUPPORTED
+	struct sctp_assoc_value assoc_val = {
+		.assoc_id = SCTP_FUTURE_ASSOC,
+		.assoc_value = val,
+	};
+	return setsockopt(fd, IPPROTO_SCTP, SCTP_AUTH_SUPPORTED, &assoc_val, sizeof(assoc_val));
+#else
+#pragma message "setsockopt(SCTP_AUTH_SUPPORTED) not supported! some SCTP features may not be available!"
+	LOGP(DLGLOBAL, LOGL_NOTICE, "Built without support for setsockopt(SCTP_AUTH_SUPPORTED), skipping\n");
+	return -ENOTSUP;
+#endif
+}
+
+static int setsockopt_sctp_asconf_supported(int fd, uint32_t val)
+{
+#ifdef SCTP_ASCONF_SUPPORTED
+	struct sctp_assoc_value assoc_val = {
+		.assoc_id = SCTP_FUTURE_ASSOC,
+		.assoc_value = val,
+	};
+	return setsockopt(fd, IPPROTO_SCTP, SCTP_ASCONF_SUPPORTED, &assoc_val, sizeof(assoc_val));
+#else
+#pragma message "setsockopt(SCTP_ASCONF_SUPPORTED) not supported! some SCTP features may not be available!"
+	LOGP(DLGLOBAL, LOGL_NOTICE, "Built without support for setsockopt(SCTP_ASCONF_SUPPORTED), skipping\n");
+	return -ENOTSUP;
+#endif
+}
+
+static int setsockopt_sctp_initmsg(int fd, const struct osmo_sock_init2_multiaddr_pars *pars)
+{
+	if (!pars->sctp.sockopt_initmsg.num_ostreams_present &&
+	    !pars->sctp.sockopt_initmsg.max_instreams_present &&
+	    !pars->sctp.sockopt_initmsg.max_attempts_present &&
+	    !pars->sctp.sockopt_initmsg.max_init_timeo_present)
+		return 0; /* nothing to set/do */
+
+#ifdef SCTP_INITMSG
+	struct sctp_initmsg si = {0};
+	socklen_t si_len = sizeof(si);
+	int rc;
+
+	/* If at least one field not present, obtain current value from kernel: */
+	if (!pars->sctp.sockopt_initmsg.num_ostreams_present ||
+	    !pars->sctp.sockopt_initmsg.max_instreams_present ||
+	    !pars->sctp.sockopt_initmsg.max_attempts_present ||
+	    !pars->sctp.sockopt_initmsg.max_init_timeo_present) {
+		rc = getsockopt(fd, IPPROTO_SCTP, SCTP_INITMSG, &si, &si_len);
+		if (rc < 0)
+			return rc;
+	}
+
+	if (pars->sctp.sockopt_initmsg.num_ostreams_present)
+		si.sinit_num_ostreams = pars->sctp.sockopt_initmsg.num_ostreams_value;
+	if (pars->sctp.sockopt_initmsg.max_instreams_present)
+		si.sinit_max_instreams = pars->sctp.sockopt_initmsg.max_instreams_value;
+	if (pars->sctp.sockopt_initmsg.max_attempts_present)
+		si.sinit_max_attempts = pars->sctp.sockopt_initmsg.max_attempts_value;
+	if (pars->sctp.sockopt_initmsg.max_init_timeo_present)
+		si.sinit_max_init_timeo = pars->sctp.sockopt_initmsg.max_init_timeo_value;
+
+	return setsockopt(fd, IPPROTO_SCTP, SCTP_INITMSG, &si, sizeof(si));
+#else
+#pragma message "setsockopt(SCTP_INITMSG) not supported! some SCTP features may not be available!"
+	LOGP(DLGLOBAL, LOGL_NOTICE, "Built without support for setsockopt(SCTP_INITMSG), skipping\n");
+	return -ENOTSUP
+#endif
+}
+
 /*! Initialize a socket (including bind and/or connect) with multiple local or remote addresses.
  *  \param[in] family Address Family like AF_INET, AF_INET6, AF_UNSPEC
  *  \param[in] type Socket type like SOCK_DGRAM, SOCK_STREAM
@@ -717,6 +790,38 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 		   const char **local_hosts, size_t local_hosts_cnt, uint16_t local_port,
 		   const char **remote_hosts, size_t remote_hosts_cnt, uint16_t remote_port,
 		   unsigned int flags)
+{
+	return osmo_sock_init2_multiaddr2(family, type, proto, local_hosts, local_hosts_cnt, local_port,
+					      remote_hosts, remote_hosts_cnt, remote_port, flags, NULL);
+}
+
+/*! Initialize a socket (including bind and/or connect) with multiple local or remote addresses.
+ *  \param[in] family Address Family like AF_INET, AF_INET6, AF_UNSPEC
+ *  \param[in] type Socket type like SOCK_DGRAM, SOCK_STREAM
+ *  \param[in] proto Protocol like IPPROTO_TCP, IPPROTO_UDP
+ *  \param[in] local_hosts array of char pointers (strings), each containing local host name or IP address in string form
+ *  \param[in] local_hosts_cnt length of local_hosts (in items)
+ *  \param[in] local_port local port number in host byte order
+ *  \param[in] remote_host array of char pointers (strings), each containing remote host name or IP address in string form
+ *  \param[in] remote_hosts_cnt length of remote_hosts (in items)
+ *  \param[in] remote_port remote port number in host byte order
+ *  \param[in] flags flags like \ref OSMO_SOCK_F_CONNECT
+ *  \param[in] pars Extra parameters for multi-address specific protocols, such as SCTP. Can be NULL.
+ *  \returns socket file descriptor on success; negative on error
+ *
+ * This function is similar to \ref osmo_sock_init2(), but can be passed an
+ * array of local or remote addresses for protocols supporting multiple
+ * addresses per socket, like SCTP (currently only one supported). This function
+ * should not be used by protocols not supporting this kind of features, but
+ * rather \ref osmo_sock_init2() should be used instead.
+ * See \ref osmo_sock_init2() for more information on flags and general behavior.
+ *
+ * pars: If "pars" parameter is passed to the function, sctp.version shall be set to 0.
+ */
+int osmo_sock_init2_multiaddr2(uint16_t family, uint16_t type, uint8_t proto,
+		   const char **local_hosts, size_t local_hosts_cnt, uint16_t local_port,
+		   const char **remote_hosts, size_t remote_hosts_cnt, uint16_t remote_port,
+		   unsigned int flags, struct osmo_sock_init2_multiaddr_pars *pars)
 
 {
 	struct addrinfo *res_loc[OSMO_SOCK_MAX_ADDRS], *res_rem[OSMO_SOCK_MAX_ADDRS];
@@ -736,6 +841,9 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 	if (proto != IPPROTO_SCTP)
 		return -ENOTSUP;
 
+	if (pars && pars->sctp.version != 0)
+		return -EINVAL;
+
 	if ((flags & (OSMO_SOCK_F_BIND | OSMO_SOCK_F_CONNECT)) == 0) {
 		LOGP(DLGLOBAL, LOGL_ERROR, "invalid: you have to specify either "
 			"BIND or CONNECT flags\n");
@@ -776,7 +884,11 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 	if (((flags & OSMO_SOCK_F_BIND) && (flags & OSMO_SOCK_F_CONNECT)) &&
 	    !addrinfo_has_in6addr_any((const struct addrinfo **)res_loc, local_hosts_cnt) &&
 	    (loc_has_v4addr != rem_has_v4addr || loc_has_v6addr != rem_has_v6addr)) {
-		LOGP(DLGLOBAL, LOGL_ERROR, "Invalid v4 vs v6 in local vs remote addresses\n");
+		LOGP(DLGLOBAL, LOGL_ERROR, "Invalid v4 vs v6 in local vs remote addresses: "
+		     "local:%s%s remote:%s%s\n",
+		     loc_has_v4addr ? " v4" : "", loc_has_v6addr ? " v6" : "",
+		     rem_has_v4addr ? " v4" : "", rem_has_v6addr ? " v6" : ""
+		     );
 		rc = -EINVAL;
 		goto ret_freeaddrinfo;
 	}
@@ -794,15 +906,61 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 		rc = setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR,
 				&on, sizeof(on));
 		if (rc < 0) {
+			int err = errno;
 			multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
 			LOGP(DLGLOBAL, LOGL_ERROR,
 			     "cannot setsockopt socket:"
 			     " %s:%u: %s\n",
 			     strbuf, local_port,
-			     strerror(errno));
+			     strerror(err));
 			goto ret_close;
 		}
 
+		if (pars && pars->sctp.sockopt_auth_supported.set) {
+			/* RFC 5061 4.2.7: ASCONF also requires AUTH feature. */
+			rc = setsockopt_sctp_auth_supported(sfd, pars->sctp.sockopt_auth_supported.value);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_AUTH_SUPPORTED) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_auth_supported.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some features such as Peer Primary Address won't be available
+				 * unless configured system-wide through sysctl */
+			}
+		}
+
+		if (pars && pars->sctp.sockopt_asconf_supported.set) {
+			rc = setsockopt_sctp_asconf_supported(sfd, pars->sctp.sockopt_asconf_supported.value);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_ASCONF_SUPPORTED) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_asconf_supported.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some features such as Peer Primary Address won't be available
+				 * unless configured system-wide through sysctl */
+			}
+		}
+
+		if (pars && pars->sctp.sockopt_initmsg.set) {
+			rc = setsockopt_sctp_initmsg(sfd, pars);
+			if (rc < 0) {
+				int err = errno;
+				multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
+				LOGP(DLGLOBAL, LOGL_ERROR,
+				     "cannot setsockopt(SCTP_INITMSG) socket: %s:%u: %s\n",
+				     strbuf, local_port, strerror(err));
+				if (pars->sctp.sockopt_initmsg.abort_on_failure)
+					goto ret_close;
+				/* do not fail, some parameters will be left as the global default */
+			}
+		}
+
 		/* Build array of addresses taking first entry for each host.
 		   TODO: Ideally we should use backtracking storing last used
 		   indexes and trying next combination if connect() fails .*/
@@ -817,9 +975,10 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 
 		rc = sctp_bindx(sfd, (struct sockaddr *)addrs_buf, local_hosts_cnt, SCTP_BINDX_ADD_ADDR);
 		if (rc == -1) {
+			int err = errno;
 			multiaddr_snprintf(strbuf, sizeof(strbuf), local_hosts, local_hosts_cnt);
 			LOGP(DLGLOBAL, LOGL_NOTICE, "unable to bind socket: %s:%u: %s\n",
-			     strbuf, local_port, strerror(errno));
+			     strbuf, local_port, strerror(err));
 			rc = -ENODEV;
 			goto ret_close;
 		}
@@ -839,9 +998,10 @@ int osmo_sock_init2_multiaddr(uint16_t family, uint16_t type, uint8_t proto,
 
 		rc = sctp_connectx(sfd, (struct sockaddr *)addrs_buf, remote_hosts_cnt, NULL);
 		if (rc != 0 && errno != EINPROGRESS) {
+			int err = errno;
 			multiaddr_snprintf(strbuf, sizeof(strbuf), remote_hosts, remote_hosts_cnt);
 			LOGP(DLGLOBAL, LOGL_ERROR, "unable to connect socket: %s:%u: %s\n",
-				strbuf, remote_port, strerror(errno));
+				strbuf, remote_port, strerror(err));
 			rc = -ENODEV;
 			goto ret_close;
 		}
@@ -1393,7 +1553,7 @@ int osmo_sock_unix_init(uint16_t type, uint8_t proto,
 
 	sfd = socket(AF_UNIX, type, proto);
 	if (sfd < 0)
-		return -1;
+		return -errno;
 
 	if (flags & OSMO_SOCK_F_CONNECT) {
 		rc = connect(sfd, (struct sockaddr *)&local, namelen);
@@ -1413,13 +1573,13 @@ int osmo_sock_unix_init(uint16_t type, uint8_t proto,
 	rc = osmo_sock_init_tail(sfd, type, flags);
 	if (rc < 0) {
 		close(sfd);
-		sfd = -1;
+		sfd = rc;
 	}
 
 	return sfd;
 err:
 	close(sfd);
-	return -1;
+	return -errno;
 }
 
 /*! Initialize a unix domain socket and fill \ref osmo_fd
@@ -1546,6 +1706,11 @@ int osmo_sock_get_name_buf(char *str, size_t str_len, int fd)
 	char portbuf_l[6], portbuf_r[6];
 	int rc;
 
+	if (fd < 0) {
+		osmo_strlcpy(str, "<error-bad-fd>", str_len);
+		return -EBADF;
+	}
+
 	/* get local */
 	if ((rc = osmo_sock_get_ip_and_port(fd, hostbuf_l, sizeof(hostbuf_l), portbuf_l, sizeof(portbuf_l), true))) {
 		osmo_strlcpy(str, "<error-in-getsockname>", str_len);
