@@ -351,7 +351,9 @@ int gsm48_encode_bearer_cap(struct msgb *msg, int lv_only,
 		lv[i++] = (bcap->data.parity & 7) |
 			  ((bcap->data.interm_rate & 3) << 5);
 		/* octet 6c */
-		lv[i] = 0x80 | (bcap->data.modem_type & 0x1f);
+		lv[i] = 0x80 |
+			((bcap->data.transp & 3) << 5) |
+			(bcap->data.modem_type & 0x1f);
 		break;
 	default:
 		return -EINVAL;
@@ -369,7 +371,7 @@ int gsm48_encode_bearer_cap(struct msgb *msg, int lv_only,
 /*! Decode TS 04.08 Call Control Capabilities IE (10.5.4.5a)
  *  \param[out] Caller-provided memory for decoded CC capabilities
  *  \param[in] lv Length-Value of IE
- *  \retursns 0 on success; negative on error */
+ *  \returns 0 on success; negative on error */
 int gsm48_decode_cccap(struct gsm_mncc_cccap *ccap, const uint8_t *lv)
 {
 	uint8_t in_len = lv[0];
@@ -825,7 +827,7 @@ int gsm48_decode_ssversion(struct gsm_mncc_ssversion *ssv,
 {
 	uint8_t in_len = lv[0];
 
-	if (in_len < 1 || in_len < sizeof(ssv->info))
+	if (in_len < 1 || in_len > sizeof(ssv->info))
 		return -EINVAL;
 
 	memcpy(ssv->info, lv + 1, in_len);
