@@ -14,10 +14,6 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 #include <errno.h>
 #include <unistd.h>
@@ -103,7 +99,7 @@ static void osmo_usb_added_cb(int fd, short events, void *user_data)
 static void osmo_usb_removed_cb(int fd, void *user_data)
 {
 	struct osmo_fd *ofd = osmo_fd_get_by_fd(fd);
-	if (!fd)
+	if (!ofd)
 		return;
 	osmo_fd_unregister(ofd);
 	talloc_free(ofd);
@@ -541,7 +537,8 @@ libusb_device_handle *osmo_libusb_open_claim_interface(void *ctx, libusb_context
 		addr = libusb_get_device_address(*dev);
 		path = osmo_libusb_dev_get_path_buf(pathbuf, sizeof(pathbuf), *dev);
 		if ((ifm->addr && addr == ifm->addr) ||
-		    (strlen(ifm->path) && !strcmp(path, ifm->path))) {
+		    (strlen(ifm->path) && !strcmp(path, ifm->path)) ||
+		    (!ifm->addr && !strlen(ifm->path) && !list[1] /* only one device */)) {
 			rc = libusb_open(*dev, &usb_devh);
 			if (rc < 0) {
 				fprintf(stderr, "Cannot open device: %s\n", libusb_error_name(rc));
@@ -739,11 +736,15 @@ int osmo_libusb_get_ep_addrs(libusb_device_handle *devh, unsigned int if_num,
 int osmo_libusb_init(libusb_context **pluctx)
 {
 	libusb_context *luctx = NULL;
+	const struct libusb_pollfd **pfds;
+
 	int rc;
 
 	rc = libusb_init(pluctx);
-	if (rc != 0)
+	if (rc != 0) {
+		LOGP(DLUSB, LOGL_ERROR, "Error initializing libusb: %s\n", libusb_strerror(rc));
 		return rc;
+	}
 
 	if (pluctx)
 		luctx = *pluctx;
@@ -754,6 +755,17 @@ int osmo_libusb_init(libusb_context **pluctx)
 
 	libusb_set_pollfd_notifiers(luctx, osmo_usb_added_cb, osmo_usb_removed_cb, luctx);
 
+	/* get the initial file descriptors which were created even before during libusb_init() */
+	pfds = libusb_get_pollfds(luctx);
+	if (pfds) {
+		const struct libusb_pollfd **pfds2 = pfds;
+		const struct libusb_pollfd *pfd;
+		/* synthesize 'add' call-backs. not sure why libusb doesn't do that by itself? */
+		for (pfd = *pfds2; pfd; pfd = *++pfds2)
+			osmo_usb_added_cb(pfd->fd, pfd->events, luctx);
+		libusb_free_pollfds(pfds);
+	}
+
 	return 0;
 }
 
