@@ -15,16 +15,12 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
- *
  */
 
 #include <stdlib.h>
 #include <string.h>
 
-#include "../../config.h"
+#include "config.h"
 
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/logging.h>
@@ -131,7 +127,7 @@ DEFUN(enable_logging,
 
 	conn = (struct telnet_connection *) vty->priv;
 	if (conn->dbg) {
-		vty_out(vty, "Logging already enabled.%s", VTY_NEWLINE);
+		vty_out(vty, "%% Logging already enabled.%s", VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -158,7 +154,7 @@ struct log_target *osmo_log_vty2tgt(struct vty *vty)
 
 	conn = (struct telnet_connection *) vty->priv;
 	if (!conn->dbg)
-		vty_out(vty, "Logging was not enabled.%s", VTY_NEWLINE);
+		vty_out(vty, "%% Logging was not enabled.%s", VTY_NEWLINE);
 
 	return conn->dbg;
 }
@@ -372,12 +368,12 @@ DEFUN(logging_level,
 	int level = log_parse_level(argv[1]);
 
 	if (level < 0) {
-		vty_out(vty, "Invalid level `%s'%s", argv[1], VTY_NEWLINE);
+		vty_out(vty, "%% Invalid level '%s'%s", argv[1], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
 	if (category < 0) {
-		vty_out(vty, "Invalid category `%s'%s", argv[0], VTY_NEWLINE);
+		vty_out(vty, "%% Invalid category '%s'%s", argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -595,7 +591,7 @@ gDEFUN(cfg_description, cfg_description_cmd,
 	char **dptr = vty->index_sub;
 
 	if (!dptr) {
-		vty_out(vty, "vty->index_sub == NULL%s", VTY_NEWLINE);
+		vty_out(vty, "%% vty->index_sub == NULL%s", VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -616,7 +612,7 @@ gDEFUN(cfg_no_description, cfg_no_description_cmd,
 	char **dptr = vty->index_sub;
 
 	if (!dptr) {
-		vty_out(vty, "vty->index_sub == NULL%s", VTY_NEWLINE);
+		vty_out(vty, "%% vty->index_sub == NULL%s", VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -834,9 +830,31 @@ DEFUN(cfg_log_gsmtap, cfg_log_gsmtap_cmd,
 	RET_WITH_UNLOCK(CMD_SUCCESS);
 }
 
+DEFUN(cfg_no_log_gsmtap, cfg_no_log_gsmtap_cmd,
+	"no log gsmtap [HOSTNAME]",
+	NO_STR LOG_STR "Logging via GSMTAP\n"
+	"Host name to send the GSMTAP logging to (UDP port 4729)\n")
+{
+	const char *hostname = argc ? argv[0] : "127.0.0.1";
+	struct log_target *tgt;
+
+	log_tgt_mutex_lock();
+	tgt = log_target_find(LOG_TGT_TYPE_GSMTAP, hostname);
+	if (tgt == NULL) {
+		vty_out(vty, "%% Unable to find GSMTAP log target for %s%s",
+			hostname, VTY_NEWLINE);
+		RET_WITH_UNLOCK(CMD_WARNING);
+	}
+
+	log_target_destroy(tgt);
+
+	RET_WITH_UNLOCK(CMD_SUCCESS);
+}
+
 DEFUN(cfg_log_stderr, cfg_log_stderr_cmd,
-	"log stderr",
-	LOG_STR "Logging via STDERR of the process\n")
+	"log stderr [blocking-io]",
+	LOG_STR "Logging via STDERR of the process\n"
+	"Use blocking, synchronous I/O\n")
 {
 	struct log_target *tgt;
 
@@ -852,6 +870,11 @@ DEFUN(cfg_log_stderr, cfg_log_stderr_cmd,
 		log_add_target(tgt);
 	}
 
+	if (argc > 0 && !strcmp(argv[0], "blocking-io"))
+		log_target_file_switch_to_stream(tgt);
+	else
+		log_target_file_switch_to_wqueue(tgt);
+
 	vty->index = tgt;
 	vty->node = CFG_LOG_NODE;
 
@@ -878,8 +901,9 @@ DEFUN(cfg_no_log_stderr, cfg_no_log_stderr_cmd,
 }
 
 DEFUN(cfg_log_file, cfg_log_file_cmd,
-	"log file .FILENAME",
-	LOG_STR "Logging to text file\n" "Filename\n")
+	"log file FILENAME [blocking-io]",
+	LOG_STR "Logging to text file\n" "Filename\n"
+	"Use blocking, synchronous I/O\n")
 {
 	const char *fname = argv[0];
 	struct log_target *tgt;
@@ -889,13 +913,18 @@ DEFUN(cfg_log_file, cfg_log_file_cmd,
 	if (!tgt) {
 		tgt = log_target_create_file(fname);
 		if (!tgt) {
-			vty_out(vty, "%% Unable to create file `%s'%s",
+			vty_out(vty, "%% Unable to create file '%s'%s",
 				fname, VTY_NEWLINE);
 			RET_WITH_UNLOCK(CMD_WARNING);
 		}
 		log_add_target(tgt);
 	}
 
+	if (argc > 1 && !strcmp(argv[1], "blocking-io"))
+		log_target_file_switch_to_stream(tgt);
+	else
+		log_target_file_switch_to_wqueue(tgt);
+
 	vty->index = tgt;
 	vty->node = CFG_LOG_NODE;
 
@@ -904,7 +933,7 @@ DEFUN(cfg_log_file, cfg_log_file_cmd,
 
 
 DEFUN(cfg_no_log_file, cfg_no_log_file_cmd,
-	"no log file .FILENAME",
+	"no log file FILENAME",
 	NO_STR LOG_STR "Logging to text file\n" "Filename\n")
 {
 	const char *fname = argv[0];
@@ -913,7 +942,7 @@ DEFUN(cfg_no_log_file, cfg_no_log_file_cmd,
 	log_tgt_mutex_lock();
 	tgt = log_target_find(LOG_TGT_TYPE_FILE, fname);
 	if (!tgt) {
-		vty_out(vty, "%% No such log file `%s'%s",
+		vty_out(vty, "%% No such log file '%s'%s",
 			fname, VTY_NEWLINE);
 		RET_WITH_UNLOCK(CMD_WARNING);
 	}
@@ -979,7 +1008,10 @@ static int config_write_log_single(struct vty *vty, struct log_target *tgt)
 		return 1;
 		break;
 	case LOG_TGT_TYPE_STDERR:
-		vty_out(vty, "log stderr%s", VTY_NEWLINE);
+		if (tgt->tgt_file.wqueue)
+			vty_out(vty, "log stderr%s", VTY_NEWLINE);
+		else
+			vty_out(vty, "log stderr blocking-io%s", VTY_NEWLINE);
 		break;
 	case LOG_TGT_TYPE_SYSLOG:
 #ifdef HAVE_SYSLOG_H
@@ -990,7 +1022,10 @@ static int config_write_log_single(struct vty *vty, struct log_target *tgt)
 #endif
 		break;
 	case LOG_TGT_TYPE_FILE:
-		vty_out(vty, "log file %s%s", tgt->tgt_file.fname, VTY_NEWLINE);
+		if (tgt->tgt_file.wqueue)
+			vty_out(vty, "log file %s%s", tgt->tgt_file.fname, VTY_NEWLINE);
+		else
+			vty_out(vty, "log file %s blocking-io%s", tgt->tgt_file.fname, VTY_NEWLINE);
 		break;
 	case LOG_TGT_TYPE_STRRB:
 		vty_out(vty, "log alarms %zu%s",
@@ -1110,6 +1145,23 @@ DEFUN(vty_logp,
 	int category = log_parse_category(argv[0]);
 	int level = log_parse_level(argv[1]);
 	char *str = argv_concat(argv, argc, 2);
+
+	if (level < 0) {
+		vty_out(vty, "%% Invalid level '%s'%s", argv[1], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (category < 0) {
+		vty_out(vty, "%% Invalid category '%s'%s", argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	/* Properly handle library specific sub-systems */
+	if ((unsigned int) category >= osmo_log_info->num_cat_user) {
+		category -= osmo_log_info->num_cat_user - 1;
+		category *= -1;
+	}
+
 	LOGP(category, level, "%s\n", str);
 	return CMD_SUCCESS;
 }
@@ -1144,7 +1196,7 @@ static void gen_vty_logp_cmd_strs(struct cmd_element *cmd)
 
 /*! Register logging related commands to the VTY. Call this once from
  *  your application if you want to support those commands. */
-void logging_vty_add_cmds()
+void logging_vty_add_cmds(void)
 {
 	install_lib_element_ve(&enable_logging_cmd);
 	install_lib_element_ve(&disable_logging_cmd);
@@ -1214,4 +1266,5 @@ void logging_vty_add_cmds()
 	install_lib_element(CONFIG_NODE, &cfg_log_systemd_journal_cmd);
 	install_lib_element(CONFIG_NODE, &cfg_no_log_systemd_journal_cmd);
 	install_lib_element(CONFIG_NODE, &cfg_log_gsmtap_cmd);
+	install_lib_element(CONFIG_NODE, &cfg_no_log_gsmtap_cmd);
 }
