@@ -1,5 +1,5 @@
 /*
- * (C) 2009-2010 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2009-2022 by Harald Welte <laforge@gnumonks.org>
  * (C) 2009-2014 by Holger Hans Peter Freyther
  * (C) 2015      by sysmocom - s.f.m.c. GmbH
  * All Rights Reserved
@@ -16,16 +16,12 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
- *
  */
 
 #include <stdlib.h>
 #include <string.h>
 
-#include "../../config.h"
+#include "config.h"
 
 #include <osmocom/vty/command.h>
 #include <osmocom/vty/buffer.h>
@@ -37,11 +33,13 @@
 #include <osmocom/core/stats.h>
 #include <osmocom/core/counter.h>
 #include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/stats_tcp.h>
 
 #define CFG_STATS_STR "Configure stats sub-system\n"
 #define CFG_REPORTER_STR "Configure a stats reporter\n"
 
 #define SHOW_STATS_STR "Show statistical values\n"
+#define SKIP_ZERO_STR "Skip items with total count zero\n"
 
 #define STATS_STR "Stats related commands\n"
 
@@ -393,27 +391,60 @@ DEFUN(cfg_stats_interval, cfg_stats_interval_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_tcp_stats_interval, cfg_tcp_stats_interval_cmd,
+	"stats-tcp interval <0-65535>",
+	CFG_STATS_STR "Set the tcp socket stats polling interval\n"
+	"Interval in seconds (0 disables the polling interval)\n")
+{
+	int rc;
+	int interval = atoi(argv[0]);
+	rc = osmo_stats_tcp_set_interval(interval);
+	if (rc < 0) {
+		vty_out(vty, "%% Unable to set interval: %s%s",
+			strerror(-rc), VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_tcp_stats_batch_size, cfg_tcp_stats_batch_size_cmd,
+	"stats-tcp batch-size <1-65535>",
+	CFG_STATS_STR "Set the number of tcp sockets that are processed per stats polling interval\n"
+	"Number of sockets per interval\n")
+{
+	osmo_tcp_stats_config->batch_size = atoi(argv[0]);
+	return CMD_SUCCESS;
+}
+
 DEFUN(show_stats,
       show_stats_cmd,
-      "show stats",
-      SHOW_STR SHOW_STATS_STR)
+      "show stats [skip-zero]",
+      SHOW_STR SHOW_STATS_STR SKIP_ZERO_STR)
 {
-	vty_out_statistics_full(vty, "");
+	bool skip_zero = false;
+	if (argc > 0)
+		skip_zero = true;
+
+	vty_out_statistics_full2(vty, "", skip_zero);
 
 	return CMD_SUCCESS;
 }
 
 DEFUN(show_stats_level,
       show_stats_level_cmd,
-      "show stats level (global|peer|subscriber)",
+      "show stats level (global|peer|subscriber) [skip-zero]",
       SHOW_STR SHOW_STATS_STR
       "Set the maximum group level\n"
       "Show global groups only\n"
       "Show global and network peer related groups\n"
-      "Show global, peer, and subscriber groups\n")
+      "Show global, peer, and subscriber groups\n" SKIP_ZERO_STR)
 {
 	int level = get_string_value(stats_class_strs, argv[0]);
-	vty_out_statistics_partial(vty, "", level);
+	bool skip_zero = false;
+	if (argc > 1)
+		skip_zero = true;
+	vty_out_statistics_partial2(vty, "", level, skip_zero);
 
 	return CMD_SUCCESS;
 }
@@ -576,23 +607,32 @@ DEFUN(show_stats_asciidoc_table,
 	return CMD_SUCCESS;
 }
 
+struct rctr_vty_ctx {
+	struct vty *vty;
+	bool skip_zero;
+};
+
 static int rate_ctr_group_handler(struct rate_ctr_group *ctrg, void *sctx_)
 {
-	struct vty *vty = sctx_;
+	struct rctr_vty_ctx *sctx = sctx_;
+	struct vty *vty = sctx->vty;
 	vty_out(vty, "%s %u", ctrg->desc->group_description, ctrg->idx);
 	if (ctrg->name != NULL)
 		vty_out(vty, " (%s)", ctrg->name);
 	vty_out(vty, ":%s", VTY_NEWLINE);
-	vty_out_rate_ctr_group_fmt(vty, "%25n: %10c (%S/s %M/m %H/h %D/d) %d", ctrg);
+	vty_out_rate_ctr_group_fmt2(vty, "%25n: %10c (%S/s %M/m %H/h %D/d) %d", ctrg, sctx->skip_zero);
 	return 0;
 }
 
 DEFUN(show_rate_counters,
       show_rate_counters_cmd,
-      "show rate-counters",
-      SHOW_STR "Show all rate counters\n")
+      "show rate-counters [skip-zero]",
+      SHOW_STR "Show all rate counters\n" SKIP_ZERO_STR)
 {
-	rate_ctr_for_each_group(rate_ctr_group_handler, vty);
+	struct rctr_vty_ctx rctx = { .vty = vty, .skip_zero = false };
+	if (argc > 0)
+		rctx.skip_zero = true;
+	rate_ctr_for_each_group(rate_ctr_group_handler, &rctx);
 	return CMD_SUCCESS;
 }
 
@@ -631,6 +671,11 @@ static int config_write_stats_reporter(struct vty *vty, struct osmo_stats_report
 	case OSMO_STATS_REPORTER_LOG:
 		type = "log";
 		break;
+	default:
+		/* don't try to save unknown stats reporters to the VTY.  Imagine some
+		 * application registering a new application specific stats reporter that
+		 * this VTY code knows nothing about! */
+		return 0;
 	}
 
 	vty_out(vty, "stats reporter %s", type);
@@ -681,6 +726,10 @@ static int config_write_stats(struct vty *vty)
 	struct osmo_stats_reporter *srep;
 
 	vty_out(vty, "stats interval %d%s", osmo_stats_config->interval, VTY_NEWLINE);
+	if (osmo_tcp_stats_config->interval != TCP_STATS_DEFAULT_INTERVAL)
+		vty_out(vty, "stats-tcp interval %d%s", osmo_tcp_stats_config->interval, VTY_NEWLINE);
+	if (osmo_tcp_stats_config->batch_size != TCP_STATS_DEFAULT_BATCH_SIZE)
+		vty_out(vty, "stats-tcp batch-size %d%s", osmo_tcp_stats_config->batch_size, VTY_NEWLINE);
 
 	/* Loop through all reporters */
 	llist_for_each_entry(srep, &osmo_stats_reporter_list, list)
@@ -693,7 +742,7 @@ static int config_write_stats(struct vty *vty)
  *  Call this once during your application initialization if you would
  *  like to have stats VTY commands enabled.
  */
-void osmo_stats_vty_add_cmds()
+void osmo_stats_vty_add_cmds(void)
 {
 	install_lib_element_ve(&show_stats_cmd);
 	install_lib_element_ve(&show_stats_level_cmd);
@@ -703,6 +752,8 @@ void osmo_stats_vty_add_cmds()
 	install_lib_element(CONFIG_NODE, &cfg_stats_reporter_log_cmd);
 	install_lib_element(CONFIG_NODE, &cfg_no_stats_reporter_log_cmd);
 	install_lib_element(CONFIG_NODE, &cfg_stats_interval_cmd);
+	install_lib_element(CONFIG_NODE, &cfg_tcp_stats_interval_cmd);
+	install_lib_element(CONFIG_NODE, &cfg_tcp_stats_batch_size_cmd);
 
 	install_node(&cfg_stats_node, config_write_stats);
 
