@@ -66,6 +66,7 @@
 #include <osmocom/vty/command.h>
 #include <osmocom/vty/buffer.h>
 #include <osmocom/core/talloc.h>
+#include <osmocom/core/timer.h>
 #include <osmocom/core/utils.h>
 
 #ifndef MAXPATHLEN
@@ -207,6 +208,12 @@ static void vty_auth(struct vty *vty)
 	}
 }
 
+void vty_flush(struct vty *vty)
+{
+	if (vty->obuf)
+		buffer_flush_all(vty->obuf, vty->fd);
+}
+
 /*! Close a given vty interface. */
 void vty_close(struct vty *vty)
 {
@@ -332,6 +339,25 @@ int vty_out_newline(struct vty *vty)
 	return 0;
 }
 
+/*! calculates the time difference of a give timespec to the current time
+ *  and prints in a human readable format (days, hours, minutes, seconds).
+ */
+int vty_out_uptime(struct vty *vty, const struct timespec *starttime)
+{
+	struct timespec now;
+	struct timespec uptime;
+
+	osmo_clock_gettime(CLOCK_MONOTONIC, &now);
+	timespecsub(&now, starttime, &uptime);
+
+	int d = uptime.tv_sec / (3600 * 24);
+	int h = uptime.tv_sec / 3600 % 24;
+	int m = uptime.tv_sec / 60 % 60;
+	int s = uptime.tv_sec % 60;
+
+	return vty_out(vty, "%dd %dh %dm %ds", d, h, m, s);
+}
+
 /*! return the current index of a given VTY */
 void *vty_current_index(struct vty *vty)
 {
@@ -459,6 +485,7 @@ static int vty_command(struct vty *vty)
 
 static const char telnet_backward_char = 0x08;
 static const char telnet_space_char = ' ';
+static const char telnet_escape_char = 0x1B;
 
 /* Basic function to write buffer to vty. */
 static void vty_write(struct vty *vty, const char *buf, size_t nbytes)
@@ -858,6 +885,19 @@ static void vty_down_level(struct vty *vty)
 	vty->cp = 0;
 }
 
+/* When '^L' is typed, clear all lines above the current one. */
+static void vty_clear_screen(struct vty *vty)
+{
+	vty_out(vty, "%c%s%c%s",
+		telnet_escape_char,
+		"[2J", /* Erase Screen */
+		telnet_escape_char,
+		"[H" /* Cursor Home */
+	);
+	vty_prompt(vty);
+	vty_redraw_line(vty);
+}
+
 /* When '^Z' is received from vty, move down to the enable mode. */
 static void vty_end_config(struct vty *vty)
 {
@@ -1402,6 +1442,9 @@ int vty_read(struct vty *vty)
 		case CONTROL('K'):
 			vty_kill_line(vty);
 			break;
+		case CONTROL('L'):
+			vty_clear_screen(vty);
+			break;
 		case CONTROL('N'):
 			vty_next_line(vty);
 			break;
@@ -1462,9 +1505,13 @@ int vty_read(struct vty *vty)
 	return 0;
 }
 
-/* Read up configuration file */
-static int
-vty_read_file(FILE *confp, void *priv)
+/* Read up configuration from a file stream */
+/*! Read up VTY configuration from a file stream
+ *  \param[in] confp file pointer of the stream for the configuration file
+ *  \param[in] priv private data to be passed to \ref vty_read_file
+ *  \returns Zero on success, non-zero on error
+ */
+int vty_read_config_filep(FILE *confp, void *priv)
 {
 	int ret;
 	struct vty *vty;
@@ -1870,7 +1917,7 @@ int vty_read_config_file(const char *file_name, void *priv)
 	if (!cfile)
 		return -ENOENT;
 
-	rc = vty_read_file(cfile, priv);
+	rc = vty_read_config_filep(cfile, priv);
 	fclose(cfile);
 
 	host_config_set(file_name);
